@@ -1,6 +1,6 @@
 (* Otags reloaded
  * 
- * Hendrik Tews Copyright (C) 2010
+ * Hendrik Tews Copyright (C) 2010 - 2012
  * 
  * This file is part of "Otags reloaded".
  * 
@@ -19,7 +19,7 @@
  * along with "Otags reloaded". If not, see
  * <http://www.gnu.org/licenses/>.
  * 
- * $Id: otags.ml,v 1.12 2010-08-27 06:57:39 tews Exp $
+ * $Id: otags.ml,v 1.19 2012-01-14 21:31:39 tews Exp $
  * 
  * main module with main function
  * 
@@ -40,7 +40,6 @@ exception Skip_entry
 
 let parser_error_message loc msg =
   prerr_endline (Loc.to_string loc);
-  prerr_string "Camlp4 parse error: ";
   prerr_endline msg
 
 
@@ -114,7 +113,7 @@ let parse_file_internally unit file =
       (try
 	 Sig_ast(pa.parse_interf ~directive_handler start_loc stream)
        with
-	 | Camlp4_parsing_error(loc, msg) ->
+	 | Otags_parsing_error(loc, msg) ->
 	   parser_error_message loc msg;
 	   Tags.empty_sig_ast
       )
@@ -129,7 +128,7 @@ let parse_file_internally unit file =
           *)
 	 Struct_ast ast
        with
-	 | Camlp4_parsing_error(loc, msg) ->
+	 | Otags_parsing_error(loc, msg) ->
 	   parser_error_message loc msg;
 	   Tags.empty_str_ast
       )
@@ -160,10 +159,15 @@ let process_file tagfun unit file =
     | false -> (parse_file_externally unit file, open_input_file file)
   in
   tagfun.start_unit file ic;
-  Tags.generate_tags ic tagfun.write_tag comp_unit;
+  (try
+     Tags.generate_tags tagfun.write_tag comp_unit
+   with
+     | Otags_parsing_error(loc, msg) ->
+       parser_error_message loc msg
+  );
   tagfun.finish_unit();
-  close_in ic
-
+  (* ic is closed by one of the tagfun's, at latest with finish_unit *)
+  ()
 
 let is_directory f explicitly_listed =
   try
@@ -178,11 +182,11 @@ let is_directory f explicitly_listed =
 
 
 (* process file or directory f, generating output for tags_out_ref.
- * If subdir_ref_option <> None than process_entry has been called
+ * If subdir_ref_option <> None then process_entry has been called
  *      recursively from process_directory. In this case it must append 
  *      detected directories to the reference in subdir_ref_option. These 
  *      directories will be processed by process_directory later.
- * If subdir_ref_option = None than process_entry has been called
+ * If subdir_ref_option = None then process_entry has been called
  *      from somewhere higher up. In this case it must invoke 
  *      process_directory 
  *      when it detects an directory (and option -r was present).
@@ -200,7 +204,10 @@ let rec process_entry tagfun f subdir_ref_option =
 	(* process_directory tagfun tags_out f *)
 	r := f :: !r
   else if ((subdir_ref_option = None) && (not !silent)) || !verbose
-  then Printf.eprintf "Skip unrecognized file %s\n" f
+  then 
+    Printf.eprintf
+      "File \"%s\", line 1\nSkip file because extension is not recognized\n"
+      f
   else ()
 
 and process_directory tagfun subdir =
@@ -222,6 +229,10 @@ and process_directory tagfun subdir =
 	     subdir_ref_option	
 	 with
 	   | Skip_entry -> ()
+	   | e -> 
+	     Printf.eprintf "Escaping exception during processing %s\n"
+	       (Filename.concat subdir entry);
+	     raise e	
 	)
       | None ->
 	not_finished := false
@@ -379,7 +390,7 @@ let usage_message =
     Sys.argv.(0)
 
 
-let run_action tag_fun = function
+let run_action tagfun = function
   | Clear_parser_list -> clear_parser_list () 
   | Add_parser parser_name -> add_parser parser_name
   | Reset_parser_list -> reset_parser_list ()
@@ -389,10 +400,13 @@ let run_action tag_fun = function
   | Process_file(file, unit_option) -> 
     try
       match unit_option with
-	| None -> process_entry tag_fun file None
-	| Some unit -> process_file tag_fun unit file
+	| None -> process_entry tagfun file None
+	| Some unit -> process_file tagfun unit file
     with
       | Skip_entry -> ()
+      | e -> 
+	Printf.eprintf "Escaping exception during processing %s\n" file;
+	raise e	  
 
 
 let main () =
@@ -449,7 +463,7 @@ let main_ex () =
 	| _ -> ()
       );
       prerr_endline "";
-      if !verbose then begin
+      if Printexc.backtrace_status() then begin
 	prerr_string backtrace;
 	prerr_endline 
 	  "\n\
